#########################################################################
#             GNUHEALTH HOSPITAL MANAGEMENT - GTK CLIENT                #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2004-2008 Tiny SPRL
# SPDX-FileCopyrightText: 2007-2009 Lorenzo Gil Sanchez
# SPDX-FileCopyrightText: 2007-2013 Bertrand Chenal
# SPDX-FileCopyrightText: 2007-2021 Cédric Krier
# SPDX-FileCopyrightText: 2008-2011 Udo Spallek>
# SPDX-FileCopyrightText: 2008-2011 Virtual things - Preisler & Spallek GbR
# SPDX-FileCopyrightText: 2008-2021 B2CK SPRL
# SPDX-FileCopyrightText: 2010-2021 Nicolas Évrard
# SPDX-FileCopyrightText: 2011-2012 Rodrigo Hübner
# SPDX-FileCopyrightText: 2012-2013 Antoine Smolders
# SPDX-FileCopyrightText: 2017-2023 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2017-2023 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2020-2021 Maxime Richez
# SPDX-FileCopyrightText: 2020-2021 SALUC SA
#
# SPDX-License-Identifier: GPL-3.0-or-later


# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from .main import Action

__all__ = [Action]
